let flow = require('lodash/flow');
const Sequelize = require('sequelize');
const fs = require('fs');
const isNil = require('lodash/isNil');
const isFunction = require('lodash/isFunction');
const isArray = require('lodash/isArray');
const async = require('async');

const PromiseFactory = require('logged-promise');
//const config = require('../config.js');
//const logger = require('../helpers/logger').logger;

const dbs = {};
// @todo 
// multidb
function getFlows(db, serviceInit, entityName) {

  function optionSetIncludeModel(options) {
    options = options || {};
    options.opt = options.opt || {};

    if (options.opt.include && options.opt.include.length > 0) {
      options.opt.include.forEach((value) => {
        if (!value || !value.modelName) throw new Error('optionSetIncludeModel: options.opt.include[].modelName is null or undefined');
        value.model = db[value.modelName];

        /* @todo: burası sınırlandırılacak! şimdilik sadece priceSpec kullanılacak */
        if (value.include && value.include.length > 0) {
          value.include.forEach((value2) => {
            if (!value2 || !value2.modelName) throw new Error('optionSetIncludeModel: options.opt.include[].modelName is null or undefined');
            value2.model = db[value2.modelName];
          });
        }

      });
    }

    return options;
  }

  function optionSetOrderModel(options) {
    options = options || {};
    options.opt = options.opt || {};

    if (options.opt.order && options.opt.order.length > 0) {
      options.opt.order.forEach((value) => {

        if (!value[0] || !value[1]) throw new Error('optionSetOrderModel: options.opt.order[] is null or undefined');

        if (value[0].modelName === entityName)
          value.splice(0, 1);

        if (db[value[0].modelName])
          value[0] = { model: db[value[0].modelName], as: value[0].as }

      });
    }

    return options;
  }

  function optionSetIdForOrganization(options) {
    options = options || {};
    options.opt = options.opt || {};
    // eger organization cekilen entity ise ID columnindan setlensin
    if (entityName === 'Organization' && serviceInit.currentClient) {
      options.opt.where = options.opt.where || {};
      options.opt.where.id = serviceInit.currentClient.organizationId;
    }

    return options;
  }
  function optionSetOrganizationId(options) {
    options = options || {};
    options.opt = options.opt || {};
    //eger organization id modelde varsa setlensin 
    if (db[entityName].attributes.hasOwnProperty('organizationId') && serviceInit.currentClient) {
      options.opt.where = options.opt.where || {};
      options.opt.where.organizationId = serviceInit.currentClient.organizationId;
    }

    return options;
  }

  function optionSetOrganizationIdForUpdateOrDelete(options) {
    options = options || {};
    options.opt = options.opt || {};
    //eger organization id modelde varsa setlensin 
    if (db[entityName].attributes.hasOwnProperty('organizationId') && serviceInit.currentClient) {
      options.opt.where = options.opt.where || {};
      options.opt.where.organizationId = serviceInit.currentClient.organizationId;
    }

    return options;
  }

  function optionSetPrimaryKeyForUpdate(options) {
    if (isNil(options.data))
      throw new Error('optionSetPrimaryKeyForUpdate: options.data is null or undefined');
    //eger opt yoksa opt yi kendimiz yaratacagiz
    if (!options.opt || !options.opt.where) {
      //options.opt = { where: {} };
      options.opt = options.opt || {};
      options.opt.where = {};
      // eger bir opt yoksa ve gelen update datasinda PK varsa opt.where e set et
      if (options.data.hasOwnProperty(db[entityName].primaryKeyField))
        options.opt.where[db[entityName].primaryKeyField] = options.data[db[entityName].primaryKeyField];
    }

    if (!options.opt.where || !options.opt.where[db[entityName].primaryKeyField])
      throw new Error('optionSetPrimaryKeyForUpdate: options.opt.where is null or undefined or does not have primary key');

    return options;
  }

  function dataTrimForUpdate(options) {
    if (isNil(options.data))
      throw new Error('dataTrimForUpdate: options.data is null or undefined');

    let trimmedUpdateData = {};
    Object.keys(options.data).forEach((value) => {

      if (options.data[value] === null) {
        trimmedUpdateData[value] = null;
      } else if (typeof options.data[value] !== "undefined")
        trimmedUpdateData[value] = options.data[value];
    });
    options.data = trimmedUpdateData;
    return options;
  }

  function dataSetOrganizationIdForCreate(options) {
    if (db[entityName].attributes.hasOwnProperty('organizationId') && serviceInit.currentClient) {
      options.data.organizationId = serviceInit.currentClient.organizationId;
    }
    return options;
  }


  function dataSetOrganizationIdForBulkCreate(options) {
    if (db[entityName].attributes.hasOwnProperty('organizationId') && serviceInit.currentClient) {

      let organizationId = serviceInit.currentClient.organizationId;
      
      options.data.forEach((item) => {
        item.organizationId = organizationId;
      });
    }
    return options;
  }


  return {
    selectFlow: flow(optionSetOrganizationId, optionSetIdForOrganization, optionSetIncludeModel, optionSetOrderModel),
    destroyFlow: flow(optionSetOrganizationIdForUpdateOrDelete, optionSetIdForOrganization),
    updateFlow: flow(dataTrimForUpdate,
      optionSetPrimaryKeyForUpdate,
      optionSetOrganizationIdForUpdateOrDelete,
      optionSetIdForOrganization),
    createFlow: flow(dataSetOrganizationIdForCreate),
    bulkCreateFlow: flow(dataSetOrganizationIdForBulkCreate)
  };
}


const mappers = {
  selectMapper: (results) => {

    if (isNil(results)) {
      return null;
    }

    if (Array.isArray(results)) {
      let mappedList = results.map((val) => val.dataValues);
      return !mappedList || !mappedList.length || mappedList.length === 0 ? null : mappedList;
    }

    return results.dataValues;
  },
  createMapper: (result) => result.dataValues,
  bulkCreateMapper: (results) => {

    let response = [];

    if (Array.isArray(results)) {
      response = results.map((val) => val.dataValues);
    } else {
      response = results.dataValues;
    }
    return response;
  },
  updateMapper: (updateResult, factoryOptions) => {
    return {
      affectedRowCount: updateResult[0],
      instance: factoryOptions.data
    };
  },
  destroyMapper: (affectedRowCount) => {
    return { affectedRowCount: affectedRowCount };
  },
  countMapper: (rowCount) => {
    return { rowCount: rowCount };
  },
  findAndCountAllMapper: (results) => {

    if (Array.isArray(results.rows)) {
      results.rows = results.rows.map((val) => val.dataValues);
    }

    return results;
  },
};

module.exports.appStart = (logger, config, onSequelizeInit) => {
  validateLogger(logger);
  validateConfig(config);

  async.each(config.databases, (_dbConfig, asyncCb) => {
    validateConnection(_dbConfig);

    dbs[_dbConfig.namespace] = {};

    let sequelize = new Sequelize(_dbConfig.connection.database, _dbConfig.connection.user, _dbConfig.connection.password, {
      host: _dbConfig.connection.host,
      port: _dbConfig.connection.port,
      dialect: _dbConfig.connection.dialect,
      storage: _dbConfig.connection.storage,
      pool: {
        max: _dbConfig.connection.maxConnection || 60,
        min: 0,
        idle: 10000
      },
      logging: logger.debug
    });

    fs.readdir(_dbConfig.models, (err, files) => {
      if (isNil(files)) {
        return asyncCb(null);
      }

      files.forEach((file) => {
        let fileName = file.split('.')[0];
        dbs[_dbConfig.namespace][fileName] = sequelize.import(_dbConfig.models + '/' + file);
      });

      return asyncCb(null);
    });
  }, (err) => {
    if (err) {
      throw err;
    }

    return onSequelizeInit(dbs);
  });
};

module.exports.init = (serviceInit, opt) => {
  if (!serviceInit || !serviceInit.requestId) throw new Error('Invalid serviceInit in sequelize-batch-model-factory');
  return {
    get: (dbName, entityName) => {
      if (isNil(dbs[dbName])) throw new Error('Database connection not found');

      let db = dbs[dbName];
      let Promise = PromiseFactory.init(serviceInit);
      let flows = getFlows(db, serviceInit, entityName);

      let run = (factoryOptions, callback) => {
        let loggedData = opt && opt.hideLogResponse ? null : factoryOptions;
        return new Promise(loggedData, function (res, rej) {
          callback(factoryOptions.flow(factoryOptions))
            .then((result) => res(factoryOptions.mapper(result, factoryOptions)))
            .catch(rej);
        }, opt || {});
      };

      return {
        findOne: (opt) => {
          if (isNil(opt))
            throw new Error('sequelize-model-factory findOne: options is null or undefined');
          let factoryOptions = {
            mapper: mappers.selectMapper,
            flow: flows.selectFlow,
            opt: opt
          };
          return run(factoryOptions, (composed) => db[entityName].findOne(composed.opt));
        },
        findById: (id, opt) => {
          if (isNil(id))
            throw new Error('sequelize-model-factory findById: id is null or undefined');

          opt = opt || {};
          opt.where = {};

          opt.where[db[entityName].primaryKeyField] = id;

          let factoryOptions = {
            mapper: mappers.selectMapper,
            flow: flows.selectFlow,
            opt: opt
          };
          return run(factoryOptions, (composed) => db[entityName].findOne(composed.opt));
        },
        findAll: (opt) => {
          let factoryOptions = {
            mapper: mappers.selectMapper,
            flow: flows.selectFlow,
            opt: opt
          };
          return run(factoryOptions, (composed) => db[entityName].findAll(composed.opt));
        },
        findAndCountAll: (opt) => {
          let factoryOptions = {
            mapper: mappers.findAndCountAllMapper,
            flow: flows.selectFlow,
            opt: opt
          };

          return run(factoryOptions, (composed) => db[entityName].findAndCountAll(composed.opt));
        },
        update: (updateData, opt) => {

          if (isNil(updateData))
            throw new Error('sequelize-model-factory update: update data is null or undefined');

          let factoryOptions = {
            mapper: mappers.updateMapper,
            flow: flows.updateFlow,
            opt: opt,
            data: updateData
          };

          return run(factoryOptions, (composed) => db[entityName].update(composed.data, composed.opt));
        },
        destroy: (opt) => {

          if (isNil(opt))
            throw new Error('sequelize-model-factory destroy: opt is null or undefined');

          let factoryOptions = {
            mapper: mappers.destroyMapper,
            flow: flows.destroyFlow,
            opt: opt
          };
          return run(factoryOptions, (composed) => db[entityName].destroy(composed.opt));
        },
        create: (createData, opt) => {
          if (isNil(createData))
            throw new Error('sequelize-model-factory create: create data is null or undefined');

          let factoryOptions = {
            mapper: mappers.createMapper,
            flow: flows.createFlow,
            opt: opt,
            data: createData
          };
          return run(factoryOptions, (composed) => db[entityName].create(composed.data, composed.opt));
        },
        count: (opt) => {
          if (isNil(opt))
            throw new Error('sequelize-model-factory count: opt is null or undefined');

          let factoryOptions = {
            mapper: mappers.countMapper,
            flow: flows.selectFlow,
            opt: opt
          };
          return run(factoryOptions, (composed) => db[entityName].count(composed.opt));
        },
        bulkCreate: (createData, opt) => {

          if (isNil(createData))
            throw new Error('sequelize-model-factory bulkCreate: create data is null or undefined');

          if (!Array.isArray(createData))
            throw new Error('sequelize-model-factory bulkCreate: create data is not array');

          let factoryOptions = {
            mapper: mappers.bulkCreateMapper,
            flow: flows.bulkCreateFlow,
            opt: opt,
            data: createData
          };
          return run(factoryOptions, (composed) => db[entityName].bulkCreate(composed.data, composed.opt));
        }
      };
    }
  };
};

function validateConnection(config) {
  if (isNil(config)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: config is null or undefined');
  if (isNil(config.namespace)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: config.namespace is null or undefined');  
  if (isNil(config.connection)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: config.connection is null or undefined');
  if (isNil(config.connection.database)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: config.connection.database is null or undefined');
  if (isNil(config.connection.user)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: config.connection.user is null or undefined');
  if (isNil(config.connection.password)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: config.connection.password is null or undefined');
  if (isNil(config.connection.host)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: config.connection.host is null or undefined');
  if (isNil(config.connection.port)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: config.connection.port is null or undefined');
  if (isNil(config.connection.dialect)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: config.connection.dialect is null or undefined');
}

const validateLogger = (logger) => {
  if (isNil(logger)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: logger is null or undefined');
  if (isNil(logger.debug)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: logger.debug is not a function');
}

const validateConfig = (config) => {
  if (isNil(config)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: config is null or undefined');
  if (isNil(config.databases) || !isArray(config.databases)) throw new Error('sequelize-model-factory validate: Invalid appStart parameters: config.databases is not an array');
};

